"""
Utilities for generating Markdown.
"""
from pathlib import Path


_EXT_TO_HINT: dict[str, str] = {
    # scripting & languages
    ".py": "python",
    ".js": "javascript",
    ".ts": "typescript",
    ".java": "java",
    ".c": "c",
    ".cpp": "cpp",
    ".cc": "cpp",
    ".cxx": "cpp",
    ".h": "cpp",
    ".hpp": "cpp",
    ".cs": "csharp",
    ".rb": "ruby",
    ".go": "go",
    ".rs": "rust",
    ".swift": "swift",
    ".kt": "kotlin",
    ".scala": "scala",
    ".dart": "dart",
    ".php": "php",
    ".pl": "perl",
    ".pm": "perl",
    ".lua": "lua",
    # web & markup
    ".html": "html",
    ".htm": "html",
    ".css": "css",
    ".scss": "scss",
    ".less": "less",
    ".json": "json",
    ".xml": "xml",
    ".yaml": "yaml",
    ".yml": "yaml",
    ".toml": "toml",
    ".ini": "ini",
    ".csv": "csv",
    ".md": "markdown",
    ".rst": "rest",
    # shell & config
    ".sh": "bash",
    ".bash": "bash",
    ".zsh": "bash",
    ".fish": "bash",
    ".ps1": "powershell",
    ".dockerfile": "dockerfile",
    # build & CI
    ".makefile": "makefile",
    ".mk": "makefile",
    "CMakeLists.txt": "cmake",
    "Dockerfile": "dockerfile",
    ".gradle": "groovy",
    ".travis.yml": "yaml",
    # data & queries
    ".sql": "sql",
    ".graphql": "graphql",
    ".proto": "protobuf",
    ".yara": "yara",
}


def syntax_hint(file_path: str | Path) -> str:
    """
    Return a syntax highlighting hint for markdown code snippet
    based on the file's extension or name.

    This can be used to annotate code blocks for rendering with syntax highlighting,
    e.g., using Markdown-style code blocks: ```<syntax_hint>\n<code>\n```.

    Args:
      file_path (str | Path): Path to the file.

    Returns:
      str: A syntax identifier suitable for code highlighting (e.g., 'python', 'json').
    """
    p = Path(file_path)
    name = p.name
    # Check full filename first (e.g., Dockerfile, CMakeLists.txt)
    if name in _EXT_TO_HINT:
        return _EXT_TO_HINT[name]
    ext = p.suffix.lower()
    if not ext:
        return ""
    return _EXT_TO_HINT.get(ext, ext.lstrip("."))



"""
Gito CI Deployment

Generates and deploys CI workflow files for automatic AI-powered code reviews.

Supported platforms:
  - GitHub Actions (.github/workflows/)
  - GitLab CI (.gitlab/ci/)

The command:
  1. Detects your Git provider (GitHub/GitLab)
  2. Prompts for LLM configuration (provider, model)
  3. Creates workflow files from templates
  4. Optionally commits and pushes to a dedicated branch
  5. Provides instructions for secrets setup

Usage:
  gito deploy [--api-type TYPE] [--commit] [--rewrite] [--to-branch NAME]

Aliases: init, connect, ci
"""
import logging
from pathlib import Path

import typer
import yaml
import microcore as mc
from microcore import ApiType, ui, utils
from git import Repo, GitCommandError
from rich.panel import Panel
from rich.console import Console

from ..core import get_base_branch
from ..cli_base import app
from ..gh_api import gh_api
from ..utils.package_metadata import version
from ..utils.git_platform.platform_types import PlatformType, identify_git_platform
from ..utils.git_platform.github import get_gh_create_pr_link, get_gh_secrets_link
from ..utils.git_platform.gitlab import (
    get_gitlab_access_tokens_link,
    get_gitlab_create_mr_link,
    get_gitlab_secrets_link,
)
from ..utils.git import get_cwd_repo_or_fail
from ..utils.cli import logo


def merge_gitlab_configs(
    file: Path,
    vars: dict  # vars reserved for future use / other merges
) -> str:
    """Merge GitLab CI configuration files."""
    # Read existing config or start with empty dict
    if file.exists():
        with open(file, 'r') as f:
            config = yaml.safe_load(f) or {}
    else:
        config = {}

    # Ensure 'stages' exists and add 'review' if not present
    if 'stages' not in config:
        config['stages'] = []
    if 'review' not in config['stages']:
        config['stages'].insert(0, 'review')  # Insert at beginning

    # Ensure 'include' exists and add the local include if not present
    if 'include' not in config:
        config['include'] = []

    # Handle case where include might be a single item (not a list)
    if not isinstance(config['include'], list):
        config['include'] = [config['include']]

    # Check if the local include already exists
    new_include = {'local': '.gitlab/ci/gito-code-review.yml'}
    include_exists = any(
        'gito-code-review.yml' in str(item.get('local', ''))
        if isinstance(item, dict) else 'gito-code-review.yml' in str(item or '')
        for item in config['include']
    )

    if not include_exists:
        config['include'].append(new_include)

    return yaml.dump(config, default_flow_style=False, sort_keys=False, indent=2)


GIT_PROVIDER_WORKFLOWS = {
    PlatformType.GITHUB: dict(
        code_review=dict(
            path=Path(".github/workflows/gito-code-review.yml"),
            template="workflows/github/gito-code-review.yml.j2",
        ),
        react_to_comments=dict(
            path=Path(".github/workflows/gito-react-to-comments.yml"),
            template="workflows/github/gito-react-to-comments.yml.j2",
        ),
    ),
    PlatformType.GITLAB: dict(
        code_review=dict(
            path=Path(".gitlab/ci/gito-code-review.yml"),
            template="workflows/gitlab/gito-code-review.yml.j2",
        ),
        gitlab_ci=dict(
            path=Path(".gitlab-ci.yml"),
            template="workflows/gitlab/.gitlab-ci.yml.j2",
            merge_function=merge_gitlab_configs,
        ),
    )
}


def _show_intro(console: Console):
    """Show introduction message for deploy command."""
    def num(n):
        return f"[green][dim][[/dim][{n:02d}][dim]][/dim][/green]"
    console.print(Panel(
        title="CI Setup",
        renderable=(
            " [bold]Wiring myself into pipelines...[/bold]\n"
            f" [green][dim]⟩[/dim]⟩⟩ INTEGRATION SEQUENCE ⟩⟩[dim]⟩[/dim] [/green] \n"
            f" {num(1)} [bold]C[/bold]onfigure language model\n"
            f" {num(2)} [bold]W[/bold]rite workflow files \n"
            f" {num(3)} [bold]C[/bold]ommit [dim]&[/dim] push to dedicated branch "
            f"[dim][[/dim]optional[dim]][/dim]       \n"  # trailing spaces to align with Logo
            f" {num(4)} [bold]G[/bold]uide you through secrets configuration"
        ),
        border_style="green",
        expand=False,
    ))


@app.command(
    name="deploy",
    help="\bCreate and deploy Gito workflows to your CI pipeline for automatic code reviews."
         "\nRun this command from your repository root."
         "\naliases: init, connect, ci"
)
@app.command(name="init", hidden=True)
@app.command(name="connect", hidden=True)
@app.command(name="ci", hidden=True)
def deploy(
    api_type: ApiType = typer.Option(None, help="LLM API type (interactive if omitted)"),
    commit: bool = typer.Option(None, help="Commit and push changes"),
    rewrite: bool = typer.Option(False, help="Overwrite existing configuration"),
    to_branch: str = typer.Option(
        default="gito-ci",
        help="Branch name for new PR containing Gito CI workflows"
    ),
    token: str = typer.Option(
        "", help="GitHub token (or set GITHUB_TOKEN env var)"
    ),
    model: str = typer.Option(
        None,
        help=(
            "Language model to use "
            "(interactive if omitted; \"default\" selects the recommended model)"
        ),
    )
) -> bool:
    """Deploy Gito to repository's CI pipeline for automatic code reviews."""
    print(logo())
    repo: Repo = get_cwd_repo_or_fail()
    console = Console()
    _show_intro(console)

    git_platform_type: PlatformType | None = identify_git_platform(repo)
    if not git_platform_type:
        ui.error("No supported Git provider detected.")
        if ui.ask_yn("Choose Git provider manually?"):
            git_platform_type = ui.ask_choose("Choose your Git provider", list(PlatformType))
        else:
            return False
    if git_platform_type not in GIT_PROVIDER_WORKFLOWS:
        ui.error(
            f"Git provider {ui.bright(git_platform_type)} is not supported "
            "for automatic deployment yet.\n"
            "Please create CI workflows manually."
        )
        return False
    workflow_files = GIT_PROVIDER_WORKFLOWS[git_platform_type]
    for file_config in workflow_files.values():
        file = file_config["path"]
        if file.exists():
            message = f"Gito CI workflow already exists at {utils.file_link(file)}"
            if rewrite:
                ui.warning(message)
            else:
                message += "\nUse --rewrite to replace existing files."
                ui.error(message)
                return False

    # configure LLM
    api_type, secret_name, model = _configure_llm(api_type, model)

    # generate workflow files from templates
    major, minor, *_ = version().split(".")
    template_vars = dict(
        model=model,
        api_type=api_type,
        secret_name=secret_name,
        major=major,
        minor=minor,
        ApiType=ApiType,
        remove_indent=True,
    )
    created_files = []
    for key, file_config in workflow_files.items():
        file: Path = file_config["path"]
        file.parent.mkdir(parents=True, exist_ok=True)
        if file.exists() and "merge_function" in file_config:
            ui.warning("Merging Gito CI workflow into existing file:", utils.file_link(file))
            merge_function = file_config["merge_function"]
            content = merge_function(file, vars=template_vars)
        else:
            template: str = file_config["template"]
            content = mc.tpl(template, **template_vars)
        if not content.endswith("\n"):
            content = content.rstrip() + "\n"
        file.write_text(content, encoding='utf-8')
        created_files.append(file)
    print(
        mc.ui.green("Gito CI workflows have been created.\n"),
        *[f"  - {mc.utils.file_link(file)}\n" for file in created_files]
    )

    # commit and push
    ui.warning('[!] Please review created files before proceeding.')

    need_to_commit = commit is True or commit is None and mc.ui.ask_yn(
        f"Commit & push CI workflows to a {mc.ui.green(to_branch)} branch?"
    )
    is_committed = False
    is_pushed = False

    if need_to_commit:
        try:
            active_branch_name = repo.active_branch.name
        except TypeError:
            active_branch_name = ""
        if active_branch_name != to_branch:
            repo.git.checkout("-b", to_branch)
        repo.git.add([str(file) for file in created_files])
        is_committed = _try_commit_workflow_changes(repo)
        if is_committed:
            is_pushed = _try_push_branch(repo, to_branch)
            if is_pushed:
                if git_platform_type == PlatformType.GITHUB:
                    try:
                        api = gh_api(repo=repo, token=token)
                        base = get_base_branch(repo).split('/')[-1]
                        logging.info(f"Creating PR {ui.green(to_branch)} -> {ui.yellow(base)}...")
                        res = api.pulls.create(
                            head=to_branch,
                            base=base,
                            title="Add Gito CI workflows",
                        )
                        print(f"Pull request #{res.number} created successfully:\n{res.html_url}")
                    except Exception as e:
                        mc.ui.error(f"Failed to create pull request automatically: {e}")
                        create_pr_link = get_gh_create_pr_link(repo, to_branch)
                        if create_pr_link:
                            details = f":\n[link]{create_pr_link}[/link]"
                        else:
                            details = "."
                        console.print(Panel(
                            title="Next step",
                            renderable=(
                                f"Please create a PR from '{to_branch}' "
                                f"to your main branch and merge it{details}"
                            ),
                            border_style="yellow",
                            expand=False,
                        ))
                elif git_platform_type == PlatformType.GITLAB:
                    create_pr_link = get_gitlab_create_mr_link(repo, to_branch)
                    if create_pr_link:
                        details = f":\n[link]{create_pr_link}[/link]"
                    else:
                        details = "."
                    console.print(Panel(
                        title="Next step",
                        renderable=(
                            f"Please create a Merge Request from branch '{to_branch}' "
                            f"to your main branch and merge it{details}"
                        ),
                        border_style="yellow",
                        expand=False,
                    ))
                else:
                    console.print(Panel(
                        title="Next step",
                        renderable=f"Please merge branch named '{to_branch}' to your main branch.",
                        border_style="yellow",
                        expand=False,
                    ))
    if not need_to_commit or not is_committed:
        console.print(Panel(
            title="Next step: Deliver CI workflows to the repository",
            renderable=(
                "Commit and push created CI workflow files to your main repository branch "
                "to activate Gito."
            ),
            border_style="yellow",
            expand=False,
        ))

    _show_create_secrets_instructions(console, git_platform_type, repo, secret_name)
    return True


def _try_commit_workflow_changes(repo: Repo) -> bool:
    """
    Try to commit workflow changes.
    Prints success or error message.
    Args:
        repo (Repo): Git repository.
    Returns:
        bool: True if commit was successful, False otherwise.
    """
    try:
        repo.git.commit("-m", "Add Gito CI workflows")
        print(ui.green("Changes committed."))
        return True
    except GitCommandError as e:
        if "nothing added" in str(e):
            ui.error("Failed to commit changes: nothing was added")
        else:
            ui.error(f"Failed to commit changes: {e}")
        return False


def _try_push_branch(repo: Repo, branch: str) -> bool:
    """
    Try to push branch to origin.
    Prints success or error message.
    Args:
        repo (Repo): Git repository.
        branch (str): Branch name.
    Returns:
        bool: True if push was successful, False otherwise.
    """
    try:
        repo.git.push("origin", branch)
        print(ui.green(
            f"Changes pushed to {ui.bright}{ui.white}{branch}{ui.reset}{ui.green} branch."
        ))
        return True
    except GitCommandError as e:
        ui.error(f"Failed to push changes: {e}")
        return False


def _configure_llm(
    api_type: str | ApiType | None,
    model: str | None = None,
) -> tuple[ApiType, str, str]:
    """
    Configure LLM.
    Args:
        api_type (str | ApiType | None): API type as string.
        model (str | None): Model name.
    Returns:
        tuple[ApiType, str, str]: (api_type, secret_name, default_model
    """
    api_types = {
        ApiType.ANTHROPIC: "Anthropic",
        ApiType.OPENAI: "OpenAI",
        ApiType.GOOGLE: "Google",
    }
    model_proposals = {
        ApiType.ANTHROPIC: {
            "claude-opus-4-5": f"Claude Opus 4.5 {ui.dim}(most capable)",
            "claude-sonnet-4-5": f"Claude Sonnet 4.5 {ui.dim}(balanced)",
            "claude-haiku-4-5": f"Claude Haiku 4.5 {ui.dim}(cheapest but less capable)",
        },
        ApiType.OPENAI: {
            "gpt-5.2": f"GPT-5.2 {ui.dim} (recommended)",
            "gpt-5.1": "GPT-5.1",
            "gpt-5": "GPT-5",
            "gpt-5-mini": "GPT-5 Mini",
        },
        ApiType.GOOGLE: {
            "gemini-2.5-pro": "Gemini 2.5 Pro",
            "gemini-2.5-flash": "Gemini 2.5 Flash",
            "gemini-3-pro-preview": f"Gemini 3 Pro Preview {ui.dim}(rate limited)",
            "gemini-3-flash-preview": f"Gemini 3 Flash Preview {ui.dim}(rate limited)",
        }
    }
    if not api_type:
        api_type = mc.ui.ask_choose(
            "Which language model API should Gito use?",
            api_types,
        )

    if api_type not in api_types:
        orig_value = api_type
        api_type = str(api_type).lower()
        if api_type == "openai":
            api_type = ApiType.OPENAI
        if api_type not in api_types:
            ui.error(f"Unsupported API type: {orig_value}")
            raise typer.Exit(2)
    secret_names = {
        ApiType.ANTHROPIC: "ANTHROPIC_API_KEY",
        ApiType.OPENAI: "OPENAI_API_KEY",
        ApiType.GOOGLE: "GOOGLE_API_KEY",
    }
    default_models = {
        ApiType.ANTHROPIC: "claude-sonnet-4-5",
        ApiType.OPENAI: "gpt-5.2",
        ApiType.GOOGLE: "gemini-2.5-pro",
    }
    use_default_model = model == "default"
    if not model or use_default_model:
        model = default_models.get(api_type, "")
        if use_default_model and not model:
            ui.error(f"No default model for API type: {api_type}")
        if not use_default_model or not model:
            if api_type in model_proposals:
                model = ui.ask_choose(
                    "Select a model",
                    model_proposals[api_type],
                    default=default_models[api_type],
                )
            else:
                model = ui.ask_non_empty("Enter the model name")

    return api_type, secret_names[api_type], model


def _show_create_secrets_instructions(
    console: Console,
    git_platform: PlatformType,
    repo: Repo,
    secret_name: str
):
    """Show instructions to create secrets in the repository."""
    details = ""
    secrets = f"  {secret_name}\n"
    title = "Add your LLM API key to repository secrets"
    if git_platform == PlatformType.GITHUB:
        if secrets_url := get_gh_secrets_link(repo):
            details = f"\n\nAdd it here:  [link]{secrets_url}[/link]"
    elif git_platform == PlatformType.GITLAB:
        title = "Add LLM API key and GitLab access token to CI/CD variables"
        details = (
            "\n\nAdd it in your GitLab project settings under "
            "Settings → CI/CD → Variables."
        )
        if secrets_url := get_gitlab_secrets_link(repo):
            details += f"\n[link]{secrets_url}[/link]"
        secrets += (
            "  GITLAB_ACCESS_TOKEN [dim]— Project Access Token "
            "with 'api' scope and 'Reporter' role\n"
            "    Create: Settings → Access Tokens[/dim]\n"
        )
        if manage_tokens_url := get_gitlab_access_tokens_link(repo):
            secrets += f"    [link]{manage_tokens_url}[/link]\n"

    console.print(Panel(
        title=f"Final step: {title}",
        renderable=(
            f"[bold yellow]Required[/bold yellow]\n"
            f"{secrets}"
            f"\n"
            f"[bold dim]Optional — Issue trackers[/bold dim]\n"
            f"  LINEAR_API_KEY, JIRA_URL, JIRA_USER, JIRA_TOKEN{details}"
        ),
        border_style="green",
        expand=False,
    ))
    if git_platform == PlatformType.GITLAB:
        console.print(Panel(
            title="Variable Settings",
            renderable=(
                "☑ Mask variable   ☐ Protect variable (uncheck, or MRs won't have access)\n"
                "\n"
                "Public repos: enable \"Require approval\" for fork pipelines"
                " in CI/CD settings to prevent secret leaks."
                "\n"
                "Docs: [link]https://docs.gitlab.com/ci/variables/[/link]"
            ),
            border_style="yellow",
            expand=False,
        ))



"""
Utility functions related to processing package metadata.
"""
import importlib.metadata


def version() -> str:
    """Return the current version of the gito.bot package."""
    return importlib.metadata.version("gito.bot")



"""
Checks if Git is installed; if not, provides installation instructions.
"""
import platform
import sys

from microcore import ui


def ensure_git_installed() -> None:
    """Ensures that Git is installed on the system."""
    try:
        from git import Repo  # noqa: F401
    except ImportError:
        _exit_with_git_instructions()


def _exit_with_git_instructions() -> None:
    """Exits the program with instructions to install Git."""
    cmd = _get_git_install_command()
    app_name = f"{ui.bright}Gito AI Code Reviewer{ui.reset}{ui.red}"
    msg_begin = (
        f"{ui.red}Error: {app_name} couldn't find "
        f"{ui.bright}Git{ui.reset}{ui.red} on the system."
        f"\n{ui.red}"
    )
    if cmd:
        details = f"To install {ui.bright}Git{ui.reset}{ui.red}, run: {ui.blue}{cmd}{ui.reset}"
    else:
        details = "Please install it from: https://git-scm.com"
    sys.exit(msg_begin + details)


def _get_git_install_command() -> str | None:
    """Returns the command to install Git based on the operating system."""
    system = platform.system().lower()

    if system == "windows":
        return "winget install Git.Git"
    elif system == "darwin":
        return "brew install git"
    elif system == "linux":
        try:
            with open("/etc/os-release") as f:
                os_release = f.read().lower()
        except FileNotFoundError:
            return None

        if "alpine" in os_release:
            return "apk add git"
        if "debian" in os_release or "ubuntu" in os_release:
            return "sudo apt install git"
        if "fedora" in os_release:
            return "sudo dnf install git"
        if "arch" in os_release:
            return "sudo pacman -S git"
    return None



"""Entry point for the Gito command-line interface (CLI)."""

from .git_installation_check import ensure_git_installed

ensure_git_installed()  # should be called before importing other modules that depend on Git

# flake8: noqa: E402, F401
from .cli import main

__all__ = ["main"]



"""
Gito core business logic.
"""
import os
import fnmatch
import logging
from typing import Iterable
from pathlib import Path
from functools import partial

import microcore as mc
from microcore import ui
from git import Repo, Commit
from git.exc import GitCommandError
from unidiff import PatchSet, PatchedFile
from unidiff.constants import DEV_NULL

from .context import Context
from .project_config import ProjectConfig
from .report_struct import ProcessingWarning, Report, ReviewTarget, RawIssue
from .constants import JSON_REPORT_FILE_NAME, REFS_VALUE_ALL
from .utils.cli import make_streaming_function
from .pipeline import Pipeline
from .env import Env
from .gh_api import gh_api


def review_subject_is_index(what):
    return not what or what == 'INDEX'


def is_binary_file(repo: Repo, file_path: str) -> bool:
    """
    Check if a file is binary by attempting to read it as text.
    Returns True if the file is binary, False otherwise.
    """
    try:
        # Attempt to read the file content from the repository tree
        content = repo.tree()[file_path].data_stream.read()
        # Try decoding as UTF-8; if it fails, it's likely binary
        content.decode("utf-8")
        return False
    except KeyError:
        try:
            fs_path = Path(repo.working_tree_dir) / file_path
            fs_path.read_text(encoding='utf-8')
            return False
        except FileNotFoundError:
            logging.error(f"File {file_path} not found in the repository.")
            return True
        except UnicodeDecodeError:
            return True
        except Exception as e:
            logging.error(f"Error reading file {file_path}: {e}")
            return True
    except UnicodeDecodeError:
        return True
    except Exception as e:
        logging.warning(f"Error checking if file {file_path} is binary: {e}")
        return True  # Conservatively treat errors as binary to avoid issues


def commit_in_branch(repo: Repo, commit: Commit, target_branch: str) -> bool:
    try:
        # exit code 0 if commit is ancestor of branch
        repo.git.merge_base('--is-ancestor', commit.hexsha, target_branch)
        return True
    except GitCommandError:
        pass
    return False


def get_base_branch(repo: Repo, pr: int | str = None):
    if os.getenv('GITHUB_ACTIONS'):

        # triggered from PR
        if base_ref := os.getenv('GITHUB_BASE_REF'):
            logging.info(f"Using GITHUB_BASE_REF:{base_ref} as base branch")
            return f'origin/{base_ref}'
        logging.info("GITHUB_BASE_REF is not available...")
        if pr:
            api = gh_api(repo=repo)
            pr_obj = api.pulls.get(pr)
            logging.info(
                f"Using 'origin/{pr_obj.base.ref}' as base branch "
                f"(received via GH API for PR#{pr})"
            )
            return f'origin/{pr_obj.base.ref}'

    try:
        logging.info(
            "Trying to resolve base branch from repo.remotes.origin.refs.HEAD.reference.name..."
        )
        # 'origin/main', 'origin/master', etc
        # Stopped working in github actions since 07/2025
        return repo.remotes.origin.refs.HEAD.reference.name
    except AttributeError:
        try:
            logging.info(
                "Checking if repo has 'main' or 'master' branches to use as --against branch..."
            )
            remote_refs = repo.remotes.origin.refs
            for branch_name in ['main', 'master']:
                if hasattr(remote_refs, branch_name):
                    return f'origin/{branch_name}'
        except Exception:
            pass

        logging.error("Could not determine default branch from remote refs.")
        raise ValueError("No default branch found in the repository.")


def get_diff(
    repo: Repo = None,
    what: str = None,
    against: str = None,
    use_merge_base: bool = True,
    pr: str | int = None
) -> PatchSet | list[PatchedFile]:
    repo = repo or Repo(".")
    if what == REFS_VALUE_ALL:
        what = get_base_branch(repo, pr=pr)
        # Git's canonical empty tree hash
        against = "4b825dc642cb6eb9a060e54bf8d69288fbee4904"
        use_merge_base = False
    if not against:
        # 'origin/main', 'origin/master', etc
        against = get_base_branch(repo, pr=pr)
    if review_subject_is_index(what):
        what = None  # working copy
    if use_merge_base:
        try:
            if review_subject_is_index(what):
                try:
                    current_ref = repo.active_branch.name
                except TypeError:
                    # In detached HEAD state, use HEAD directly
                    current_ref = "HEAD"
                    logging.info(
                        "Detected detached HEAD state, using HEAD as current reference"
                    )
            else:
                current_ref = what
            merge_base = repo.merge_base(current_ref or repo.active_branch.name, against)[0]
            logging.info(
                f"Merge base({ui.green(current_ref)},{ui.yellow(against)})"
                f" --> {ui.cyan(merge_base.hexsha)}"
            )
            # if branch is already an ancestor of "against", merge_base == branch ⇒ it’s been merged
            if merge_base.hexsha == repo.commit(current_ref or repo.active_branch.name).hexsha:
                # @todo: check case: reviewing working copy index in main branch #103
                logging.info(
                    f"Branch is already merged. ({ui.green(current_ref)} vs {ui.yellow(against)})"
                )
                merge_sha = repo.git.log(
                    '--merges',
                    '--ancestry-path',
                    f'{current_ref}..{against}',
                    '-n',
                    '1',
                    '--pretty=format:%H'
                ).strip()
                if merge_sha:
                    logging.info(f"Merge commit is {ui.cyan(merge_sha)}")
                    merge_commit = repo.commit(merge_sha)

                    other_merge_parent = None
                    for parent in merge_commit.parents:
                        logging.info(f"Checking merge parent: {parent.hexsha[:8]}")
                        if parent.hexsha == merge_base.hexsha:
                            logging.info(f"merge parent is {ui.cyan(parent.hexsha[:8])}, skipping")
                            continue
                        if not commit_in_branch(repo, parent, against):
                            logging.warning(f"merge parent is not in {against}, skipping")
                            continue
                        logging.info(f"Found other merge parent: {ui.cyan(parent.hexsha[:8])}")
                        other_merge_parent = parent
                        break
                    if other_merge_parent:
                        first_common_ancestor = repo.merge_base(other_merge_parent, merge_base)[0]
                        # for gito remote (feature_branch vs origin/main)
                        # the same merge base appears in first_common_ancestor again
                        if first_common_ancestor.hexsha == merge_base.hexsha:
                            if merge_base.parents:
                                first_common_ancestor = repo.merge_base(
                                    other_merge_parent, merge_base.parents[0]
                                )[0]
                            else:
                                logging.error(
                                    "merge_base has no parents, "
                                    "using merge_base as first_common_ancestor"
                                )
                        logging.info(
                            f"{what} will be compared to "
                            f"first common ancestor of {what} and {against}: "
                            f"{ui.cyan(first_common_ancestor.hexsha[:8])}"
                        )
                        against = first_common_ancestor.hexsha
                    else:
                        logging.error(f"Can't find other merge parent for {merge_sha}")
                else:
                    logging.warning(
                        f"No merge‐commit found for {current_ref!r}→{against!r}; "
                        "falling back to merge‐base diff"
                    )
            else:
                # normal case: branch not yet merged
                against = merge_base.hexsha
                logging.info(
                    f"Using merge base: {ui.cyan(merge_base.hexsha[:8])} ({merge_base.summary})"
                )
        except Exception as e:
            logging.error(f"Error finding merge base: {e}")
    logging.info(
        f"Making diff: {ui.green(what or 'INDEX')} vs {ui.yellow(against)}"
    )
    diff_content = repo.git.diff(against, what)
    diff = PatchSet.from_string(diff_content)

    # Filter out binary files
    non_binary_diff = PatchSet([])
    for patched_file in diff:
        # Check if the file is binary using the source or target file path
        file_path = (
            patched_file.target_file
            if patched_file.target_file != DEV_NULL
            else patched_file.source_file
        )
        if file_path == DEV_NULL:
            continue
        if is_binary_file(repo, file_path.removeprefix("b/")):
            logging.info(f"Skipping binary file: {patched_file.path}")
            continue
        non_binary_diff.append(patched_file)
    return non_binary_diff


def filter_diff(
    patch_set: PatchSet | Iterable[PatchedFile],
    filters: str | list[str],
    exclude: bool = False,
) -> PatchSet | Iterable[PatchedFile]:
    """
    Filter the diff files by the given fnmatch filters.
    Args:
        patch_set (PatchSet | Iterable[PatchedFile]): The diff to filter.
        filters (str | list[str]): The fnmatch patterns to filter by.
        exclude (bool): If True, inverse logic (exclude files matching the filters).
    Returns:
        PatchSet | Iterable[PatchedFile]: The filtered diff.
    """
    if not isinstance(filters, (list, str)):
        raise ValueError("Filters must be a string or a list of strings")
    if not isinstance(filters, list):
        filters = [f.strip() for f in filters.split(",") if f.strip()]
    if not filters:
        return patch_set
    files = [
        file
        for file in patch_set
        if (
            not any(fnmatch.fnmatch(file.path, pattern) for pattern in filters) if exclude
            else any(fnmatch.fnmatch(file.path, pattern) for pattern in filters)
        )
    ]
    return files


def read_file(repo: Repo, file: str, use_local_files: bool = False) -> str:
    if use_local_files:
        file_path = Path(repo.working_tree_dir) / file
        try:
            return file_path.read_text(encoding='utf-8')
        except (FileNotFoundError, UnicodeDecodeError) as e:
            logging.warning(f"Could not read file {file} from working directory: {e}")

    # Read from HEAD (committed version)
    return repo.tree()[file].data_stream.read().decode('utf-8')


def file_lines(
    repo: Repo,
    file: str,
    max_tokens: int = None,
    use_local_files: bool = False
) -> str:
    """
    Read file content and return it with line numbers.
    If max_tokens is specified, trims the content to fit within the token limit.
    Args:
        repo (Repo): The git repository.
        file (str): The file path to read.
        max_tokens (int, optional): Maximum number of tokens to return. Defaults to None.
        use_local_files (bool): Whether to read from local working directory first.
    Returns:
        str: The file content with line numbers.
    """
    text = read_file(repo=repo, file=file, use_local_files=use_local_files)
    lines = [f"{i + 1}: {line}\n" for i, line in enumerate(text.splitlines())]
    if max_tokens:
        lines, removed_qty = mc.tokenizing.fit_to_token_size(lines, max_tokens)
        if removed_qty:
            lines.append(
                f"(!) DISPLAYING ONLY FIRST {len(lines)} LINES DUE TO LARGE FILE SIZE\n"
            )
    return "".join(lines)


def read_files(repo: Repo, files: list[str], max_tokens: int = None) -> dict:
    out = dict()
    total_tokens = 0
    for file in files:
        content = read_file(repo=repo, file=file, use_local_files=True)
        total_tokens += mc.tokenizing.num_tokens_from_string(file)
        total_tokens += mc.tokenizing.num_tokens_from_string(content)
        if max_tokens and total_tokens > max_tokens:
            logging.warning(
                f"Skipping file {file} due to exceeding max_tokens limit ({max_tokens})"
            )
            continue
        out[file] = content
    return out


def make_cr_summary(ctx: Context, **kwargs) -> str:
    return (
        mc.prompt(
            ctx.config.summary_prompt,
            diff=mc.tokenizing.fit_to_token_size(ctx.diff, ctx.config.max_code_tokens)[0],
            issues=ctx.report.issues,
            pipeline_out=ctx.pipeline_out,
            env=Env,
            **ctx.config.prompt_vars,
            **kwargs,
        ).to_llm()
        if ctx.config.summary_prompt
        else ""
    )


class NoChangesInContextError(Exception):
    """
    Exception raised when there are no changes in the context to review or answer questions.
    """


def get_target_diff(
    repo: Repo,
    config: ProjectConfig,
    what: str = None,
    against: str = None,
    filters: str | list[str] = "",
    use_merge_base: bool = True,
    pr: str | int = None,
) -> PatchSet | Iterable[PatchedFile]:
    """
    Get the target diff for review or answering questions.
    Applies filtering based on the provided filters and project configuration.
    Raises NoChangesInContextError if no changes are found after filtering.
    Returns:
        PatchSet | Iterable[PatchedFile]: The filtered diff.
    """
    diff = get_diff(
        repo=repo, what=what, against=against, use_merge_base=use_merge_base, pr=pr,
    )
    diff = filter_diff(diff, filters)
    if config.exclude_files:
        diff = filter_diff(diff, config.exclude_files, exclude=True)
    if not diff:
        raise NoChangesInContextError()
    return diff


def get_target_lines(
    repo: Repo,
    config: ProjectConfig,
    diff: PatchSet | Iterable[PatchedFile],
    what: str = None,
) -> dict[str, str]:
    """
    Get the lines of code for each file in the diff.
    Returns a dictionary mapping file paths to their respective lines of code.
    """
    lines = {
        file_diff.path: (
            file_lines(
                repo,
                file_diff.path,
                config.max_code_tokens
                - mc.tokenizing.num_tokens_from_string(str(file_diff)),
                use_local_files=review_subject_is_index(what) or what == REFS_VALUE_ALL
            )
            if file_diff.target_file != DEV_NULL or what == REFS_VALUE_ALL
            else ""
        )
        for file_diff in diff
    }
    return lines


def _prepare(
    repo: Repo = None,
    what: str = None,
    against: str = None,
    filters: str | list[str] = "",
    use_merge_base: bool = True,
    pr: str | int = None,
):
    repo = repo or Repo(".")
    cfg = ProjectConfig.load_for_repo(repo)
    diff = get_target_diff(
        repo=repo,
        config=cfg,
        what=what,
        against=against,
        filters=filters,
        use_merge_base=use_merge_base,
        pr=pr,
    )
    lines = get_target_lines(repo=repo, config=cfg, diff=diff, what=what)
    return repo, cfg, diff, lines


def get_affected_code_block(repo: Repo, file: str, start_line: int, end_line: int) -> str | None:
    if not start_line or not end_line:
        return None
    try:
        if isinstance(start_line, str):
            start_line = int(start_line)
        if isinstance(end_line, str):
            end_line = int(end_line)
        lines = file_lines(repo, file, max_tokens=None, use_local_files=True)
        if lines:
            lines = [""] + lines.splitlines()
            return "\n".join(
                lines[start_line: end_line + 1]
            )
    except Exception as e:
        logging.error(
            f"Error getting affected code block for {file} from {start_line} to {end_line}: {e}"
        )
    return None


def provide_affected_code_blocks(
    issues: dict,
    repo: Repo,
    processing_warnings: list = None
):
    """
    For each issue, fetch the affected code text block
    and add it to the issue data.
    """
    for file, file_issues in issues.items():
        for issue in file_issues:
            try:
                for i in issue.get("affected_lines", []):
                    file_name = i.get("file", issue.get("file", file))
                    if block := get_affected_code_block(
                        repo,
                        file_name,
                        i.get("start_line"),
                        i.get("end_line")
                    ):
                        i["affected_code"] = block
            except Exception as e:
                logging.exception(e)
                if processing_warnings is None:
                    continue
                processing_warnings.append(
                    ProcessingWarning(
                        message=(
                            f"Error fetching affected code blocks for file {file}: {e}"
                        ),
                        file=file,
                    )
                )


def _llm_response_validator(parsed_response: list[dict]):
    """
    Validate that the LLM response is a list of dicts that can be converted to RawIssue.
    """
    if not isinstance(parsed_response, list):
        raise ValueError("Response is not a list")
    for item in parsed_response:
        if not isinstance(item, dict):
            raise ValueError("Response item is not a dict")
        RawIssue(**item)
    return True


async def review(
    target: ReviewTarget,
    repo: Repo = None,
    out_folder: str | os.PathLike | None = None,
):
    """
    Conducts a code review.
    Prints the review report to the console and saves it to a file.
    """
    try:
        repo, cfg, diff, lines = _prepare(
            repo=repo,
            what=target.what,
            against=target.against,
            filters=target.filters,
            use_merge_base=target.use_merge_base,
            pr=target.pull_request_id,
        )
    except NoChangesInContextError:
        logging.error("No changes to review")
        return

    def input_is_diff(file_diff: PatchedFile) -> bool:
        """
        In case of reviewing all changes, or added files,
        we provide full file content as input.
        Otherwise, we provide the diff and additional file lines separately.
        """
        return not target.is_full_codebase_review() and not file_diff.is_added_file

    responses = await mc.llm_parallel(
        [
            mc.prompt(
                cfg.prompt,
                input=(
                    file_diff if input_is_diff(file_diff)
                    else str(file_diff.path) + ":\n" + lines[file_diff.path]
                ),
                file_lines=lines[file_diff.path] if input_is_diff(file_diff) else None,
                **cfg.prompt_vars,
            )
            for file_diff in diff
        ],
        retries=cfg.retries,
        parse_json={"validator": _llm_response_validator},
        allow_failures=True,
    )
    processing_warnings: list[ProcessingWarning] = []
    for i, (res_or_error, file) in enumerate(zip(responses, diff)):
        if isinstance(res_or_error, Exception):
            if isinstance(res_or_error, mc.LLMContextLengthExceededError):
                message = f'File "{file.path}" was skipped due to large size: {str(res_or_error)}.'
            else:
                message = (
                    f"File {file.path} was skipped due to error: "
                    f"[{type(res_or_error).__name__}] {res_or_error}"
                )
                if not message.endswith('.'):
                    message += '.'
            processing_warnings.append(
                ProcessingWarning(
                    message=message,
                    file=file.path,
                )
            )
            responses[i] = []

    issues = {file.path: issues for file, issues in zip(diff, responses) if issues}
    provide_affected_code_blocks(issues, repo, processing_warnings)
    exec(cfg.post_process, {"mc": mc, **locals()})
    out_folder = Path(out_folder or repo.working_tree_dir)
    out_folder.mkdir(parents=True, exist_ok=True)
    report = Report(
        target=target,
        number_of_processed_files=len(diff),
        processing_warnings=processing_warnings,
    )
    report.register_issues(issues)
    ctx = Context(
        report=report,
        config=cfg,
        diff=diff,
        repo=repo,
    )
    if cfg.pipeline_steps:
        pipe = Pipeline(
            ctx=ctx,
            steps=cfg.pipeline_steps
        )
        pipe.run()
    else:
        logging.info("No pipeline steps defined, skipping pipeline execution")

    report.summary = make_cr_summary(ctx)
    report.save(file_name=out_folder / JSON_REPORT_FILE_NAME)
    report_text = report.render(cfg, Report.Format.MARKDOWN)
    text_report_path = out_folder / "code-review-report.md"
    text_report_path.write_text(report_text, encoding="utf-8")
    report.to_cli()


def answer(
    question: str,
    repo: Repo = None,
    what: str = None,
    against: str = None,
    filters: str | list[str] = "",
    use_merge_base: bool = True,
    use_pipeline: bool = True,
    prompt_file: str = None,
    pr: str | int = None,
    aux_files: list[str] = None,
) -> str | None:
    """
    Answers a question about the code changes.
    Returns the LLM response as a string.
    """
    try:
        repo, config, diff, lines = _prepare(
            repo=repo,
            what=what,
            against=against,
            filters=filters,
            use_merge_base=use_merge_base,
            pr=pr
        )
    except NoChangesInContextError:
        logging.error("No changes to review")
        return

    ctx = Context(
        repo=repo,
        diff=diff,
        config=config,
        report=Report()
    )
    if use_pipeline:
        pipe = Pipeline(
            ctx=ctx,
            steps=config.pipeline_steps
        )
        pipe.run()

    if aux_files or config.aux_files:
        aux_files_dict = read_files(
            repo,
            (aux_files or []) + config.aux_files,
            config.max_code_tokens // 2
        )
    else:
        aux_files_dict = {}

    if not prompt_file and config.answer_prompt.startswith("tpl:"):
        prompt_file = str(config.answer_prompt)[4:]

    if prompt_file:
        prompt_func = partial(mc.tpl, prompt_file)
    else:
        prompt_func = partial(mc.prompt, config.answer_prompt)
    prompt = prompt_func(
        question=question,
        diff=diff,
        all_file_lines=lines,
        pipeline_out=ctx.pipeline_out,
        aux_files=aux_files_dict,
        **config.prompt_vars,
    )
    response = mc.llm(
        prompt,
        callback=make_streaming_function() if Env.verbosity == 0 else None,
    )
    return response



import hashlib
import json
from typing import Literal

from gito.report_struct import Issue

from .report_struct import Report

GitLabSeverity = Literal["info", "minor", "major", "critical", "blocker"]

SEVERITY_MAP: dict[int, GitLabSeverity] = {
    5: "minor",  # from: Suggestion
    4: "minor",  # from: Trivial
    3: "minor",  # from: Minor
    2: "major",  # from: Major
    1: "critical",  # from: Critical
    0: "blocker",  # not applicable
}


def _generate_fingerprint(*parts: str) -> str:
    """Generate a stable fingerprint hash from the given parts."""
    content = "\0".join(parts)
    return hashlib.md5(content.encode(), usedforsecurity=False).hexdigest()


def _map_severity(issue: Issue) -> GitLabSeverity:
    """Map issue severity/confidence to GitLab severity levels."""
    if hasattr(issue, "severity") and issue.severity in SEVERITY_MAP:
        return SEVERITY_MAP[issue.severity]
    return "minor"


def convert_to_gitlab_code_quality_report(report: Report, **kwargs) -> str:
    """
    Convert a code review report into GitLab Code Quality report format.

    Args:
        report: The code review report to convert.

    Returns:
        JSON string in GitLab Code Quality format.

    See:
        https://docs.gitlab.com/ee/ci/testing/code_quality.html#implement-a-custom-tool
    """
    gitlab_report = [
        {
            "description": (
                issue.title
                + (f":\n{issue.details}" if issue.details else "")
                + (f"\n[Tags]: {', '.join(issue.tags)}" if issue.tags else "")
                + (
                    "\n[Proposed change]:"
                    f"\n```{line.syntax_hint}\n{line.proposal.strip()}\n```"
                    if line.proposal else ""
                )
            ),
            "fingerprint": _generate_fingerprint(
                file_path,
                str(line.start_line),
                str(line.end_line),
                issue.title,
            ),
            "severity": _map_severity(issue),
            "location": {
                "path": file_path,
                "lines": {
                    "begin": line.start_line,
                    "end": line.end_line,
                },
            },
        }
        for file_path, issues in report.issues.items()
        for issue in issues
        for line in issue.affected_lines
    ]

    return json.dumps(gitlab_report, indent=2)



from typing import Optional

from ..platform_types import PlatformType
from ..github import (
    get_gh_secrets_link,
    is_running_in_github_action,
    gh_ci_src_branch,
    get_gh_create_pr_link,
    get_gh_file_link,
)
from .base import BaseGitPlatform


class GitHubPlatform(BaseGitPlatform):
    type: PlatformType = PlatformType.GITHUB

    def is_running_in_ci(self) -> bool:
        return is_running_in_github_action()

    def ci_src_branch(self) -> Optional[str]:
        return gh_ci_src_branch()

    def create_pr_url(self, branch: str) -> Optional[str]:
        return get_gh_create_pr_link(self.repo_base_url, branch)

    def secrets_management_url(self) -> Optional[str]:
        return get_gh_secrets_link(self.repo_base_url)

    def file_url(
        self,
        file: str,
        branch="main",
        start_line: Optional[int] = None,
        end_line: Optional[int] = None,
    ) -> Optional[str]:
        return get_gh_file_link(
            repo_or_base_url=self.repo_base_url,
            file=file,
            branch=branch,
            start_line=start_line,
            end_line=end_line,
        )



import os
import sys
import logging

import requests
import typer

from ..cli_base import app, arg_refs
from ..issue_trackers import resolve_issue_key
from ..utils.git import get_cwd_repo_or_fail


def post_linear_comment(issue_key: str, text: str, api_key: str):
    """
    Post a comment to a Linear issue using the Linear API.
    Args:
        issue_key (str): The ID of the Linear issue to comment on.
        text (str): The comment text to post.
        api_key (str): The Linear API key for authentication.
    Returns:
        dict: The JSON response from the Linear API.
    """
    response = requests.post(
       'https://api.linear.app/graphql',
       headers={'Authorization': api_key, 'Content-Type': 'application/json'},
       json={
           'query': '''
               mutation($issueId: String!, $body: String!) {
                   commentCreate(input: {issueId: $issueId, body: $body}) {
                       comment { id }
                   }
               }
           ''',
           'variables': {'issueId': issue_key, 'body': text}
       }
    )
    return response.json()


@app.command(help="Post a comment with specified text to the associated Linear issue.")
def linear_comment(
    text: str = typer.Argument(None),
    refs: str = arg_refs(),
):
    if text is None or text == "-":
        # Read from stdin if no text provided
        text = sys.stdin.read()

    if not text or not text.strip():
        typer.echo("Error: No comment text provided.", err=True)
        raise typer.Exit(code=1)

    api_key = os.getenv("LINEAR_API_KEY")
    if not api_key:
        logging.error("LINEAR_API_KEY environment variable is not set")
        return

    repo = get_cwd_repo_or_fail()
    key = resolve_issue_key(repo)
    post_linear_comment(key, text, api_key)
    logging.info("Comment posted to Linear issue %s", key)



import os
import asyncio
import logging
import sys
import textwrap

import microcore as mc
import typer
from gito.utils.git_platform.shared import get_repo_base_web_url

from .core import (
    review,
    answer,
    get_target_diff,
    get_base_branch,
    NoChangesInContextError,
)
from .cli_base import (
    app,
    args_to_target,
    arg_refs,
    arg_what,
    arg_filters,
    arg_out,
    arg_against,
    arg_all,
    get_repo_context,
)
from .report_struct import Report, ReviewTarget
from .constants import HOME_ENV_PATH, GITHUB_MD_REPORT_FILE_NAME, REFS_VALUE_ALL
from .bootstrap import bootstrap
from .utils.cli import no_subcommand, logo
from .utils.html import remove_html_comments
from .utils.git_platform.shared import get_repo_domain_and_path
from .utils.git_platform.platform_types import PlatformType, identify_git_platform
from .project_config import ProjectConfig

from .commands.gh_post_review_comment import post_github_cr_comment
from .commands.gitlab_post_review_comment import post_gitlab_cr_comment
from .commands.linear_comment import linear_comment
# Imported for registering commands
from .commands import fix, gh_react_to_comment, repl, deploy, version  # noqa

app_no_subcommand = typer.Typer(pretty_exceptions_show_locals=False)


def main():
    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    # Help subcommand alias: if 'help' appears as first non-option arg, replace it with '--help'
    if len(sys.argv) > 1 and sys.argv[1] == "help":
        sys.argv = [sys.argv[0]] + sys.argv[2:] + ["--help"]

    if no_subcommand(app):
        bootstrap()
        app_no_subcommand()
    else:
        app()


@app.callback(
    invoke_without_command=True,
    help="\bGito is an open-source AI code reviewer that works with any language model provider."
         "\nIt detects issues in GitHub pull requests or local codebase changes"
         "—instantly, reliably, and without vendor lock-in."
)
def cli(
    ctx: typer.Context,
    verbosity: int = typer.Option(
        None,
        '--verbosity', '-v',
        show_default=False,
        help="\b"
             "Set verbosity level. Supported values: 0-3. Default: 1."
             "\n [ 0 ]: no additional output, "
             "\n [ 1 ]: normal mode, shows warnings, shortened LLM requests and logging.INFO"
             "\n [ 2 ]: verbose mode, show full LLM requests"
             "\n [ 3 ]: very verbose mode, also debug information"
    ),
    verbose: bool = typer.Option(
        default=None,
        help="\b"
             "--verbose is equivalent to -v2, "
             "\n--no-verbose is equivalent to -v0. "
             "\n(!) Can't be used together with -v or --verbosity."
    ),
):
    if verbose is not None and verbosity is not None:
        raise typer.BadParameter(
            "Please specify either --verbose or --verbosity, not both."
        )
    if verbose is not None:
        verbosity = 2 if verbose else 0
    if verbosity is None:
        verbosity = 1

    if ctx.invoked_subcommand != "setup":
        bootstrap(verbosity)


def _consider_arg_all(all: bool, refs: str, merge_base: bool) -> tuple[str, bool]:
    """
    Handle the --all option logic for commands.
    Returns:
        Updated (refs, merge_base) tuple.
    """
    if all:
        if refs and refs != REFS_VALUE_ALL:
            raise typer.BadParameter(
                "The --all option overrides the refs argument. "
                "Please remove the refs argument if you want to review all codebase."
            )
        refs = REFS_VALUE_ALL
        merge_base = False
    return refs, merge_base


@app_no_subcommand.command(name="review", help="Perform code review")
@app.command(name="review", help="Perform a code review of the target codebase changes.")
@app.command(name="run", hidden=True)
def cmd_review(
    refs: str = arg_refs(),
    what: str = arg_what(),
    against: str = arg_against(),
    filters: str = arg_filters(),
    merge_base: bool = typer.Option(default=True, help="Use merge base for comparison"),
    url: str = typer.Option("", "--url", help="Git repository URL"),
    path: str = typer.Option("", "--path", help="Git repository path"),  # @todo: implement
    post_comment: bool = typer.Option(
        default=False,
        help="Post review comment to git platform (GitHub, GitLab, etc.)"
    ),
    pr: int = typer.Option(
        default=None,
        help=textwrap.dedent("""\n
        GitHub Pull Request number or GitLab Merge Request ID to post the comment to
        (for local usage together with --post-comment,
        in the GitHub/GitLab actions PR/MR is resolved from the environment)
        """)
    ),
    out: str = arg_out(),
    all: bool = arg_all(),
):
    refs, merge_base = _consider_arg_all(all, refs, merge_base)
    _what, _against = args_to_target(refs, what, against)
    pr = pr or os.getenv("PR_NUMBER_FROM_WORKFLOW_DISPATCH")
    with (get_repo_context(url, _what) as (repo, out_folder)):
        commit_sha = repo.head.commit.hexsha
        try:
            active_branch = repo.active_branch.name
        except TypeError:
            active_branch = None
        review_target = ReviewTarget(
            git_platform_type=identify_git_platform(repo),
            repo_url=get_repo_base_web_url(repo),
            pull_request_id=str(pr) if pr else None,
            what=_what,
            against=_against,
            filters=filters,
            use_merge_base=merge_base,
            commit_sha=commit_sha,
            active_branch=active_branch,
        )
        asyncio.run(review(
            repo=repo,
            target=review_target,
            out_folder=out or out_folder,
        ))
        if post_comment:

            md_report_file = os.path.join(out or out_folder, GITHUB_MD_REPORT_FILE_NAME)
            if review_target.git_platform_type == PlatformType.GITHUB:
                try:
                    _, repo_path = get_repo_domain_and_path(repo)
                except ValueError as e:
                    logging.error(
                        "Error posting comment:\n"
                        "Could not extract GitHub repository path "
                        "from the local repository."
                    )
                    raise typer.Exit(code=1) from e
                post_github_cr_comment(
                    md_report_file=md_report_file,
                    pr=pr,
                    gh_repo=repo_path,
                )
            elif review_target.git_platform_type == PlatformType.GITLAB:
                post_gitlab_cr_comment(md_report_file=md_report_file, merge_request_iid=pr)
            else:
                msg = "Posting comments is only supported for GitHub and GitLab repositories."
                if not review_target.git_platform_type:
                    msg = f"Could not identify the Git provider for the current repository. {msg}"
                logging.error(msg)
                raise typer.Exit(code=1)


@app.command(name="ask", help="Answer questions about the target codebase changes.")
@app.command(name="answer", hidden=True)
def cmd_answer(
    question: str = typer.Argument(help="Question to ask about the codebase changes"),
    refs: str = arg_refs(),
    what: str = arg_what(),
    against: str = arg_against(),
    filters: str = arg_filters(),
    merge_base: bool = typer.Option(default=True, help="Use merge base for comparison"),
    use_pipeline: bool = typer.Option(default=True),
    post_to: str = typer.Option(
        help="Post answer to ... Supported values: linear",
        default=None,
        show_default=False
    ),
    pr: int = typer.Option(
        default=None,
        help="GitHub Pull Request number"
    ),
    aux_files: list[str] = typer.Option(
        default=None,
        help="Auxiliary files that might be helpful"
    ),
    save_to: str = typer.Option(
        help="Write the answer to the target file",
        default=None,
        show_default=False
    ),
    all: bool = arg_all(),
):
    refs, merge_base = _consider_arg_all(all, refs, merge_base)
    _what, _against = args_to_target(refs, what, against)
    pr = pr or os.getenv("PR_NUMBER_FROM_WORKFLOW_DISPATCH")
    if str(question).startswith("tpl:"):
        prompt_file = str(question)[4:]
        question = ""
    else:
        prompt_file = None
    out = answer(
        question=question,
        what=_what,
        against=_against,
        filters=filters,
        use_merge_base=merge_base,
        prompt_file=prompt_file,
        use_pipeline=use_pipeline,
        pr=pr,
        aux_files=aux_files,
    )
    if post_to == 'linear':
        logging.info("Posting answer to Linear...")
        linear_comment(remove_html_comments(out))
    if save_to:
        with open(save_to, "w", encoding="utf-8") as f:
            f.write(out)
        logging.info(f"Answer saved to {mc.utils.file_link(save_to)}")

    return out


@app.command(help="Configure LLM for local usage interactively.")
def setup():
    print(logo())
    mc.interactive_setup(HOME_ENV_PATH)


@app.command(name="report", help="Render and display code review report.")
@app.command(name="render", hidden=True)
def render(
    format: str = typer.Argument(
        default=Report.Format.CLI,
        help="Report format: md (Markdown), cli (terminal)"
    ),
    source: str = typer.Option(
        "",
        "--src",
        "--source",
        help="Source file (json) to load the report from"
    )
):
    Report.load(file_name=source).to_cli(report_format=format)


@app.command(
    help="\bList files in the changeset. "
         "\nMight be useful to check what will be reviewed if run `gito review` "
         "with current CLI arguments and options."
)
def files(
    refs: str = arg_refs(),
    what: str = arg_what(),
    against: str = arg_against(),
    filters: str = arg_filters(),
    merge_base: bool = typer.Option(default=True, help="Use merge base for comparison"),
    diff: bool = typer.Option(default=False, help="Show diff content"),
    all: bool = arg_all(),
):
    refs, merge_base = _consider_arg_all(all, refs, merge_base)
    _what, _against = args_to_target(refs, what, against)
    with get_repo_context(url=None, branch=_what) as (repo, out_folder):
        cfg = ProjectConfig.load_for_repo(repo)
        try:
            patch_set = get_target_diff(
                repo=repo,
                config=cfg,
                what=_what,
                against=_against,
                filters=filters,
                use_merge_base=merge_base,
                pr=None,
            )
        except NoChangesInContextError:
            patch_set = []

        print(
            f"Changed files: "
            f"{mc.ui.green(_what or 'INDEX')} vs "
            f"{mc.ui.yellow(_against or get_base_branch(repo))}"
            f"{' filtered by ' + mc.ui.cyan(filters) if filters else ''} --> "
            f"{mc.ui.cyan(len(patch_set))} file(s)."
        )

        for patch in patch_set:
            if patch.is_added_file:
                color = mc.ui.green
            elif patch.is_removed_file:
                color = mc.ui.red
            else:
                color = mc.ui.blue
            print(f"- {color(patch.path)}")
            if diff:
                print(mc.ui.gray(textwrap.indent(str(patch), "  ")))



import logging
import os
import re
from dataclasses import dataclass, field

import git

from .utils.git_platform.github import is_running_in_github_action, gh_ci_src_branch
from .utils.git_platform.gitlab import is_running_in_gitlab_ci


@dataclass
class IssueTrackerIssue:
    title: str = field(default="")
    description: str = field(default="")
    url: str = field(default="")


def extract_issue_key(branch_name: str, min_len=2, max_len=10) -> str | None:
    """
    Extracts an issue key from the given branch name.
    The issue key is expected to be in the format PROJECT-123,
    where PROJECT is an uppercase alphanumeric string of length between min_len and max_len,
    followed by a hyphen and a numeric identifier.
    Args:
        branch_name (str): The branch name to extract the issue key from.
        min_len (int): Minimum length of the project key part.
        max_len (int): Maximum length of the project key part.
    Returns:
        str | None: The extracted issue key, or None if not found.
    """
    boundary = r'\b|_|-|/|\\'
    pattern = fr"(?:{boundary})([A-Z][A-Z0-9]{{{min_len - 1},{max_len - 1}}}-\d+)(?:{boundary})"
    match = re.search(pattern, branch_name)
    return match.group(1) if match else None


def get_branch(repo: git.Repo):
    if is_running_in_github_action():
        return gh_ci_src_branch()
    elif is_running_in_gitlab_ci():
        # @todo: consider using gitlab_ci_src_branch(),
        # test differences with variables used there for different workflows first
        # See: https://docs.gitlab.com/ci/variables/predefined_variables/
        gitlab_ref = os.getenv('CI_COMMIT_REF_NAME')
        if gitlab_ref:
            return gitlab_ref
    try:
        branch_name = repo.active_branch.name
        return branch_name
    except Exception as e:  # @todo: specify more precise exception
        logging.error("Could not determine the active branch name: %s", e)
        return None


def resolve_issue_key(repo: git.Repo):
    branch_name = get_branch(repo)
    if not branch_name:
        logging.error("No active branch found in the repository, cannot determine issue key.")
        return None

    if not (issue_key := extract_issue_key(branch_name)):
        logging.error(f"No issue key found in branch name: {branch_name}")
        return None
    return issue_key



from typing import Optional

from git import Repo


def get_repo_domain_and_path(repo: Repo) -> tuple[str, str]:
    """
    Extract the domain and repository path from the git remote URL.
    Examples:
        git@github.com:owner/repo.git -> ('github.com', 'owner/repo')
        https://gitlab.com/owner/repo.git -> ('gitlab.com', 'owner/repo')
    Args:
        repo (git.Repo): The git repository object.
    Returns:
        tuple[str, str]: A tuple containing the domain and repository path.
    Raises:
        ValueError: If the remote URL format is unsupported.
    """
    try:
        remote_url = repo.remotes.origin.url
    except Exception as e:
        raise ValueError("Could not get remote URL from the repository.") from e
    if remote_url.endswith(".git"):
        remote_url = remote_url[:-4]
    if remote_url.startswith('git@') and ':' in remote_url:
        domain, path = remote_url[4:].split(':', 1)
        return domain, path
    elif remote_url.startswith('https://'):
        domain, path = remote_url[8:].split('/', 1)
        return domain, path
    else:
        raise ValueError("Unsupported git remote URL format")


def get_repo_base_web_url(repo: Repo) -> Optional[str]:
    """
    Get the base web URL of the repository.
    Tested / supported platforms: GitHub, GitLab.
    Args:
        repo (Repo): The git repository object.
    Returns:
        Optional[str]: The web URL of the repository.
    """
    try:
        domain, path = get_repo_domain_and_path(repo)
    except ValueError:
        return None
    return f"https://{domain}/{path}"


def get_repo_web_url(
    repo: Repo | str,
    subpath: str
) -> Optional[str]:
    """
    Get the web URL of the repository.
    Tested / supported platforms: GitHub, GitLab.
    Args:
        repo (Repo | str): The git repository object or repository base URL.
        subpath (str): Path to append to the base URL.
    Returns:
        Optional[str]: The web URL of the repository.
    """
    if isinstance(repo, Repo):
        repo = get_repo_base_web_url(repo)
    return (repo + "/" + subpath.lstrip('/')) if repo else None


def get_repo_owner_and_name(repo: Repo) -> tuple[str, str]:
    """
    Extract the repository owner and repository name.

    Returns:
        tuple[str, str]: A tuple containing the owner and repository name.
    Raises:
        ValueError: If the remote URL format is unsupported.
    """
    domain, path = get_repo_domain_and_path(repo)
    owner, repo_name = path.split('/', 1)
    return owner, repo_name



"""
String utility functions.
"""


def max_line_len(text: str) -> int:
    """
    Return the length of the longest line in the given text.
    Args:
        text (str): The input text.
    Returns:
        int: The length of the longest line.
    """
    return max((len(line) for line in text.splitlines()), default=0)


def block_wrap_lr(
    text: str,
    left: str = "",
    right: str = "",
    max_rwrap: int = 60,
    min_wrap: int = 0,
) -> str:
    """
    Wrap the given text block with specified left and right strings.
    If the maximum line length of the text is less than or equal to max_rwrap,
    the right string is added to the end of each line, padded with spaces
    to align with the longest line.
    Args:
        text (str): The input text block.
        left (str): The string to prepend to each line.
        right (str): The string to append to each line if wrapping conditions are met.
        max_rwrap (int): The maximum line length to apply right wrapping.
        min_wrap (int): The minimum line length to consider for wrapping.
    Returns:
        str: The wrapped text block.
    """
    ml = max(max_line_len(text), min_wrap)
    lines = text.splitlines()
    wrapped_lines = []
    for line in lines:
        ln = left + line
        if ml <= max_rwrap:
            ln += ' ' * (ml - len(line)) + right
        wrapped_lines.append(ln)
    return "\n".join(wrapped_lines)


def parse_refs_pair(refs: str) -> tuple[str | None, str | None]:
    SEPARATOR = '..'
    if not refs:
        return None, None
    if SEPARATOR not in refs:
        return refs, None
    what, against = refs.split(SEPARATOR, 1)
    return what or None, against or None



"""Show Gito version command."""
from ..cli_base import app
from ..env import Env


@app.command(name='version', help='Show Gito version.')
def version():
    print(Env.gito_version)
    return Env.gito_version



"""
Posting code review comments to GitLab Merge Requests.
"""
import logging
import os
from time import sleep
from typing import List, Dict, Optional

import requests
import typer
from microcore import ui

from ..cli_base import app
from ..constants import GITHUB_MD_REPORT_FILE_NAME, HTML_CR_COMMENT_MARKER
from ..project_config import ProjectConfig
from ..utils.git_platform.gitlab import is_running_in_gitlab_ci


def resolve_gl_token(token: str | None) -> Optional[str]:
    """Resolve GitLab access token from CLI flag or env var."""
    return (token or "").strip() or os.getenv("GITLAB_ACCESS_TOKEN") or os.getenv("GITLAB_TOKEN")


def require_gl_token(token: str | None) -> str:
    """Resolve GitLab access token from CLI flag or env var, or exit."""
    if token := resolve_gl_token(token):
        return token
    hint = (
        "Add it to CI/CD variables (Settings → CI/CD → Variables)."
        if is_running_in_gitlab_ci()
        else "Pass --token or set GITLAB_ACCESS_TOKEN env var."
    )
    ui.error(
        f"GitLab access token is required.\n"
        f"{hint}\n"
        "Create a Project Access Token (role: Reporter, scope: api) at:\n"
        "Settings → Access Tokens"
    )
    raise typer.Exit(1)


def _gl_base_url(base_url: Optional[str]) -> str:
    return (base_url or os.getenv("GITLAB_BASE_URL") or "https://gitlab.com").rstrip("/")


def post_gl_comment(
    project_id: str,
    merge_request_iid: int,
    token: str,
    body: str,
    base_url: Optional[str] = None,
) -> bool:
    """Create a note on a GitLab Merge Request."""
    base_url = _gl_base_url(base_url)
    url = f"{base_url}/api/v4/projects/{project_id}/merge_requests/{merge_request_iid}/notes"
    headers = {"PRIVATE-TOKEN": token}
    resp = requests.post(url, headers=headers, json={"body": body}, timeout=30)
    if resp.status_code != 201:
        logging.error(
            "Failed to post GitLab MR note: %s %s", resp.status_code, resp.text
        )
        return False
    return True


def list_gl_mr_notes(
    project_id: str,
    merge_request_iid: int,
    token: str,
    base_url: Optional[str] = None,
) -> List[Dict]:
    """List *all* notes on a GitLab MR (handles pagination)."""
    all_notes: List[Dict] = []
    page = 1
    headers = {"PRIVATE-TOKEN": token}
    base = _gl_base_url(base_url)

    while True:
        url = (
            f"{base}/api/v4/projects/{project_id}/merge_requests/{merge_request_iid}/notes"
            f"?per_page=100&page={page}"
        )
        resp = requests.get(url, headers=headers, timeout=30)
        if resp.status_code != 200:
            logging.error(
                "Failed to list GitLab MR notes: %s %s", resp.status_code, resp.text
            )
            break
        notes = resp.json() or []
        all_notes.extend(notes)
        next_page = resp.headers.get("X-Next-Page")
        if not next_page:
            break
        page = int(next_page)

    # Ensure notes are sorted chronologically by creation time
    all_notes.sort(key=lambda n: n.get("created_at", ""))
    return all_notes


def update_gl_mr_note(
    project_id: str,
    merge_request_iid: int,
    note_id: int,
    token: str,
    new_body: str,
    base_url: Optional[str] = None,
) -> bool:
    """Update a GitLab MR note with new body content."""
    url = (
        f"{_gl_base_url(base_url)}/api/v4/projects/{project_id}/merge_requests/"
        f"{merge_request_iid}/notes/{note_id}"
    )
    headers = {"PRIVATE-TOKEN": token}
    resp = requests.put(url, headers=headers, json={"body": new_body}, timeout=30)
    if resp.status_code != 200:
        logging.error(
            "Failed to update GitLab MR note %s: %s %s",
            note_id,
            resp.status_code,
            resp.text,
        )
        return False
    return True


def collapse_gl_outdated_cr_comments(
    project_id: str,
    merge_request_iid: int,
    token: Optional[str] = None,
    base_url: Optional[str] = None,
) -> None:
    """
    Collapse previous CR comments on a GitLab MR by wrapping them in <details>…</details>.

    We detect our comments using HTML_CR_COMMENT_MARKER. All but the most recent
    matching note are collapsed. GitLab doesn't support "minimize" via API for normal
    MR notes, so we edit the note bodies in-place.
    """
    logging.info(
        "Collapsing outdated comments in GitLab project %s MR !%s...",
        project_id,
        merge_request_iid,
    )
    token = require_gl_token(token)

    notes = list_gl_mr_notes(project_id, merge_request_iid, token, base_url)

    review_marker = HTML_CR_COMMENT_MARKER
    collapsed_title = "🗑️ Outdated Code Review by Gito"
    collapsed_marker = f"<summary>{collapsed_title}</summary>"

    candidates = [
        n
        for n in notes
        if (n.get("body") and (review_marker in n["body"]) and (collapsed_marker not in n["body"]))
    ]

    # Exclude the most recent matching note
    outdated = candidates[:-1] if candidates else []
    if not outdated:
        logging.info("No outdated comments found")
        return

    for n in outdated:
        note_id = n["id"]
        logging.info("Collapsing GitLab MR note %s...", note_id)
        new_body = f"<details>\n<summary>{collapsed_title}</summary>\n\n{n['body']}\n</details>"
        update_gl_mr_note(project_id, merge_request_iid, note_id, token, new_body, base_url)

    logging.info("All outdated comments collapsed successfully.")


@app.command(name="gitlab-comment", help="Leave a GitLab MR comment with the review.")
@app.command(name="post-gitlab-comment", hidden=True)
def post_gitlab_cr_comment(
    md_report_file: str = typer.Option(
        default=None,
        help=(
            "Path to the markdown review file. "
            "Gito's standard report file will be used by default."
        ),
    ),
    project_id: str = typer.Option(
        default=None, help="GitLab project ID (numeric) or URL-encoded path"
    ),
    merge_request_iid: int = typer.Option(default=None, help="Merge Request IID"),
    token: str = typer.Option(
        "", help="GitLab access token (or set GITLAB_ACCESS_TOKEN env var)"
    ),
    base_url: Optional[str] = typer.Option(
        default=None, help="GitLab base URL (default env GITLAB_BASE_URL or https://gitlab.com)"
    ),
):
    """
    Leaves a comment with the review on the current GitLab merge request.

    Requires a Project Access Token with 'api' scope.
    The default $CI_JOB_TOKEN does not have write access to merge requests.

    Examples:
      ```bash
      gito gitlab-comment \
        --token $GITLAB_ACCESS_TOKEN \
        --project-id $CI_PROJECT_ID \
        --merge-request-iid $CI_MERGE_REQUEST_IID
      ```
    """
    file = md_report_file or GITHUB_MD_REPORT_FILE_NAME
    if not os.path.exists(file):
        logging.error(f"Review file not found: {file}, comment will not be posted.")
        raise typer.Exit(4)

    with open(file, "r", encoding="utf-8") as f:
        body = f.read()

    token = require_gl_token(token)

    # Resolve project and MR IID from flags or common CI env vars
    project_id = project_id or os.getenv("CI_PROJECT_ID")
    mr_env_val = os.getenv("CI_MERGE_REQUEST_IID")
    if not merge_request_iid and mr_env_val:
        try:
            merge_request_iid = int(mr_env_val)
        except ValueError:
            pass

    if not project_id:
        logging.error("Could not resolve GitLab project_id (flag or CI_PROJECT_ID env var).")
        raise typer.Exit(3)

    if not merge_request_iid:
        logging.error(
            "Could not resolve GitLab merge_request_iid (flag or CI_MERGE_REQUEST_IID env var)."
        )
        raise typer.Exit(3)

    # Post the note
    if not post_gl_comment(project_id, merge_request_iid, token, body, base_url):
        raise typer.Exit(5)

    # Optionally collapse older comments
    config = ProjectConfig.load()
    if getattr(config, "collapse_previous_code_review_comments", False):
        sleep(1)
        collapse_gl_outdated_cr_comments(project_id, merge_request_iid, token, base_url)



import git
from ..platform_types import PlatformType
from .base import BaseGitPlatform
from .github import GitHubPlatform
from .gitlab import GitLabPlatform


IMPLEMENTATIONS = {
    PlatformType.GITHUB: GitHubPlatform,
    PlatformType.GITLAB: GitLabPlatform,
}


def get_platform_adapter(
    platform_type: PlatformType,
    repo_or_base_url: git.Repo | str,
) -> BaseGitPlatform:
    """
    Factory function to get the appropriate Git platform adapter.
    Args:
        platform_type (PlatformType): The type of Git platform.
        repo_or_base_url (git.Repo | str): The git repository or its base URL.
    Returns:
        BaseGitPlatform: An instance of the appropriate Git platform adapter.
    Raises:
        ValueError: If the platform type is unsupported or if the input type is invalid.
    """
    if isinstance(repo_or_base_url, git.Repo):
        repo_base_url = None
        repo = repo_or_base_url
    elif isinstance(repo_or_base_url, str):
        repo_base_url = repo_or_base_url
        repo = None
    else:
        raise ValueError("repo_or_base_url must be a git.Repo or str")

    adapter_class = IMPLEMENTATIONS.get(platform_type)
    if not adapter_class:
        raise ValueError(f"Unsupported platform type: {platform_type}")
    return adapter_class(repo=repo, repo_base_url=repo_base_url)


__all__ = [
    "BaseGitPlatform",
    "GitHubPlatform",
    "GitLabPlatform",
]



"""
Utilities for working with GitLab platform.
"""
import os
from typing import Optional

import git
from urllib.parse import quote

from .shared import get_repo_web_url


def is_running_in_gitlab_ci() -> bool:
    """
    Check if the code is running inside a GitLab CI/CD environment.

    Returns:
        bool: True if running in GitLab CI, False otherwise.

    See: https://docs.gitlab.com/ci/variables/predefined_variables/
    """
    return os.getenv("GITLAB_CI") == "true"


def extract_gitlab_owner_repo(repo: git.Repo) -> tuple[str, str]:
    """
    Extracts the GitLab owner and repository name.

    Returns:
        tuple[str, str]: A tuple containing the owner and repository name.
    """
    try:
        remote_url = repo.remotes.origin.url
    except Exception as e:
        raise ValueError("Could not get remote URL from the repository.") from e
    if remote_url.startswith('git@gitlab.com:'):
        # SSH format: git@gitlab.com:owner/repo.git
        repo_path = remote_url.split(':')[1].replace('.git', '')
    elif remote_url.startswith('https://gitlab.com/'):
        # HTTPS format: https://gitlab.com/owner/repo.git
        repo_path = remote_url.replace('https://gitlab.com/', '').replace('.git', '')
    else:
        raise ValueError("Unsupported remote URL format")
    parts = repo_path.rsplit('/', 1)
    if len(parts) != 2:
        raise ValueError("Unsupported gitlab repository path format")
    owner, repo_name = parts
    return owner, repo_name


def get_gitlab_create_mr_link(repo_or_base_url: git.Repo | str, branch: str) -> Optional[str]:
    """
    Return a GitLab URL to create a merge request for the given branch.
    """
    branch = quote(branch, safe='')
    return get_repo_web_url(
        repo_or_base_url, f"/-/merge_requests/new?merge_request%5Bsource_branch%5D={branch}"
    )


def gitlab_ci_src_branch() -> Optional[str]:
    """
    Get the current branch name in a GitLab CI environment.
    If not running in GitLab CI, returns None.
    """
    # See: https://docs.gitlab.com/ci/variables/predefined_variables/
    if not is_running_in_gitlab_ci():
        return None
    if "CI_MERGE_REQUEST_SOURCE_BRANCH_NAME" in os.environ:
        return os.environ["CI_MERGE_REQUEST_SOURCE_BRANCH_NAME"]
    elif "CI_COMMIT_BRANCH" in os.environ:
        return os.environ["CI_COMMIT_BRANCH"]
    return None


def get_gitlab_secrets_link(repo_or_base_url: git.Repo | str) -> Optional[str]:
    """
    Return a GitLab URL to manage secrets.
    Returns None in case of error.
    Args:
        repo_or_base_url (git.Repo | str): The git repository object or repository base URL.
    """
    return get_repo_web_url(repo_or_base_url, "/-/settings/ci_cd#js-cicd-variables-settings")


def get_gitlab_access_tokens_link(repo_or_base_url: git.Repo | str) -> Optional[str]:
    """
    Return a GitLab URL to create an access token.
    Returns None in case of error.
    """
    return get_repo_web_url(repo_or_base_url, "/-/settings/access_tokens")


def get_gitlab_file_link(
    repo_or_base_url: git.Repo | str,
    file: str,
    branch="main",
    start_line: Optional[int] = None,
    end_line: Optional[int] = None,
) -> Optional[str]:
    """
    Return a GitLab URL to view a file on a branch, optionally anchored to line numbers.
    Returns None in case of error.
    """
    branch, file = quote(branch, safe=''), quote(file, safe='/')
    url_path = f"/-/blob/{branch}/{file}?ref_type=heads"
    if start_line is not None:
        url_path += f"#L{start_line}"
        if end_line is not None and end_line != start_line:
            url_path += f"-{end_line}"
    return get_repo_web_url(repo_or_base_url, url_path)



"""
Common CLI arguments and utilities for Gito commands.
"""
import contextlib
import logging
import tempfile
from typing import Iterator

import microcore as mc
import typer
from git import Repo, InvalidGitRepositoryError

from .constants import REFS_VALUE_ALL
from .utils.string import parse_refs_pair
from .env import Env


app = typer.Typer(pretty_exceptions_show_locals=False)


def args_to_target(refs, what, against) -> tuple[str | None, str | None]:
    """Convert CLI arguments to target WHAT and AGAINST refs."""
    if refs == REFS_VALUE_ALL:
        return REFS_VALUE_ALL, None
    _what, _against = parse_refs_pair(refs)
    if _what:
        if what:
            raise typer.BadParameter(
                "You cannot specify both 'refs' <WHAT>..<AGAINST> and '--what'. Use one of them."
            )
    else:
        _what = what
    if _against:
        if against:
            raise typer.BadParameter(
                "You cannot specify both 'refs' <WHAT>..<AGAINST> and '--against'. Use one of them."
            )
    else:
        _against = against
    return _what, _against


def arg_refs() -> typer.Argument:
    return typer.Argument(
        default=None,
        help=(
            "Git refs to review, [what]..[against] (e.g., 'HEAD..HEAD~1'). "
            "If omitted, the current index (including added but not committed files) "
            "will be compared to the repository’s main branch."
        ),
    )


def arg_what() -> typer.Option:
    return typer.Option(None, "--what", "-w", help="Git ref to review")


def arg_filters() -> typer.Option:
    return typer.Option(
        "", "--filter", "-f", "--filters",
        help="""
            filter reviewed files by glob / fnmatch pattern(s),
            e.g. 'src/**/*.py', may be comma-separated
            """,
    )


def arg_out() -> typer.Option:
    return typer.Option(
        None,
        "--out", "-o", "--output",
        help="Output folder for the code review report"
    )


def arg_against() -> typer.Option:
    return typer.Option(
        None,
        "--against", "-vs", "--vs",
        help="Git ref to compare against"
    )


def arg_all() -> typer.Option:
    return typer.Option(default=False, help="Review whole codebase")


@contextlib.contextmanager
def get_repo_context(
    url: str | None,
    branch: str | None
) -> Iterator[tuple[Repo, str]]:
    """
    Context manager for handling both local and remote repositories.
    Yields a tuple of (Repo object, path to the repository)
    Args:
        url (str): URL of the remote repository. If empty, uses the local repository.
        branch (str): Branch to checkout when cloning the remote repository.
    """
    if branch == REFS_VALUE_ALL:
        branch = None
    if url:
        with tempfile.TemporaryDirectory() as temp_dir:
            logging.info(
                f"get_repo_context: "
                f"Cloning [{mc.ui.green(url)}] to {mc.utils.file_link(temp_dir)} ..."
            )
            repo = Repo.clone_from(url, branch=branch, to_path=temp_dir)
            prev_folder = Env.working_folder
            Env.working_folder = temp_dir
            try:
                yield repo, temp_dir
            finally:
                repo.close()
                Env.working_folder = prev_folder
    else:
        logging.info("get_repo_context: Using local repo...")
        try:
            repo = Repo(".")
        except InvalidGitRepositoryError:
            raise typer.BadParameter("Current folder is not a git repository.")
        try:
            yield repo, "."
        finally:
            repo.close()
